﻿#----------------------------------------------
# SQL Server Health check script - Daily
# 2013-05-01 Dean Willson
# blog: http://dean-o.blogspot.com
# twitter: @deanwillson
# email: deanowillson@gmail.com
#----------------------------------------------
# ASSUMPTIONS
# 1. SQL Connections running as account with SELECT permissions
#    to msdb on each server it runs agains
# 2. Express editions can be detected by the presence of
#    "Express Edition" in the @@Server_Version property
# 3. Servers using a weekly full and daily Differentials
#    will show up as exceptions every day except after
#    the full backup runs.
# 4. Remote TCP connections are enabled
# 5. SMTP host is can relay emails from the IP that
#    this script is executed from with the credentials
#    this script is executing as
# 6. This script is run daily in the morning
# 7. Not sure if this script is compatible with SQL 2000
#----------------------------------------------
# DISCLAIMER
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#----------------------------------------------
# Revision	by	Description
#----------------------------------------------
# 10/24/2013 v0.6.1 ddw Revise Check 4 to select max(backup_finish_date)
# 10/23/2013 v0.6.0 ddw Rewrite Check 3, 4, 5 queries
# 10/22/2013 v0.5.0 ddw Add variables for logging server and database.
#                       Add emailing from this script.
# 08/17/2013 v0.4.0 ddw add where clause to check 5 to exclude
#                       SQL Express instance checks since there
#                       is not SQL Agent for the Express Edition
# 07/30/2013 v0.3.0 ddw put try catch finally block around
#                       step 5 to fix issue. turned out to be server
#                       name field varchar(20) was too short
#                       for sqlexpress instance. changed to varchar(50)
# 06/28/2013 v0.2.0 ddw write to database table
# 06/28/2013 v0.1.0 ddw Added a check of last full database backup
#----------------------------------------------
$loggingServer = 'AdminServerName'
$loggingDatabase = 'AdminUtility'
$instanceListPath = "C:\Automating_SqlServer_Health_Checks\Instances.txt"
$NbrDaysOfJobHistory = 1
$NbrDaysWithoutBackup = 1
$NbrDaysWithoutLogBackup = 1
# change below to $false if reporting is done from another system
# such as Reporting Services
$useThisScriptToSendEmail = $true 
$emailServerHost = 'smtp.yourhost.com'
$emailFrom = 'noreply@yourhost.com'
$emailTo = 'me@yourhost.com'
$emailSubject = 'SQL Server Daily Health Check'

cls
#----------------------------------------------
# Function to insert the failed check details
#----------------------------------------------
Function InsertDailyHealthCheck ($server, $checkId, $priority, $findingsGroup, $finding, $details){
	#
	# TODO: move the connection and command outside the loop
	#
	$conn = new-object System.Data.SqlClient.SqlConnection "server=$loggingServer;database=$loggingDatabase;Integrated Security=sspi"
	$conn.Open()
	$comm = $conn.CreateCommand()
	$runDate = Get-Date -Format 'yyyy-MM-dd'

	$sqlCmd = "INSERT INTO [dbo].[SqlServerDailyHealthCheck]
           ([Server]
           ,[CheckID]
           ,[Priority]
           ,[FindingsGroup]
           ,[Finding]
           ,[Details]
		   ,[RunDate]
           )
     VALUES
           (@Server
           ,@CheckId
           ,@Priority
           ,@FindingsGroup
           ,@Finding
           ,@Details
		   ,@RunDate
          )"
	$comm.CommandText = $sqlCmd
	$comm.Parameters.AddWithValue("@Server", $server) | Out-Null
	$comm.Parameters.AddWithValue("@CheckId", $checkId) | Out-Null
	$comm.Parameters.AddWithValue("@Priority", $priority) | Out-Null
	$comm.Parameters.AddWithValue("@FindingsGroup", $findingsGroup) | Out-Null
	$comm.Parameters.AddWithValue("@Finding", $finding) | Out-Null
	$comm.Parameters.AddWithValue("@Details", $details) | Out-Null
	$comm.Parameters.AddWithValue("@RunDate", $runDate) | Out-Null

	[Void]$comm.ExecuteNonquery()
	$conn.Close()
	
}

#-----------------------------------------
# Start the health checks
#-----------------------------------------
$checkId = 0
$priority = 1

Write-Host "1) Agent Job - Failed to complete" -BackgroundColor Blue -ForegroundColor White

# Check 1
# for each server, get SQL Agent job failure info
#
$checkId = $checkId + 1
foreach ($svr in get-content $instanceListPath){
	#write-host $svr
	
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$svr;database=msdb;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT sjh.[server]
			,sj.[name]
			,CONVERT(VARCHAR(500), sjh.[message]) AS [message]
			,sjh.run_date
			,sjh.run_time
		FROM msdb..sysjobhistory sjh
		INNER JOIN msdb..sysjobs sj ON sjh.job_id = sj.job_id
		INNER JOIN (
			SELECT job_id
				,max(instance_id) maxinstanceid
			FROM msdb..sysjobhistory
			WHERE run_status NOT IN (1,4) --succeeded, in progress
			GROUP BY job_id
		) a ON sjh.job_id = a.job_id
		AND sjh.instance_id = a.maxinstanceid
		WHERE DATEDIFF(dd, CONVERT(VARCHAR(8), sjh.run_date), GETDATE()) <= $NbrDaysOfJobHistory"
	$cmd.CommandText = $sql
	$da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	[void]$da.fill($ds) 
	$cn.Close()
	$ds.Tables[0] | foreach {
		$tmpDetails = $_.name + " run date: " + $_.run_date + " " + $_.message
		# log it
		InsertDailyHealthCheck $svr $checkId $priority 'Agent Job' 'Failed to complete' $tmpDetails 

	}
	
}#end for each server

# Check 2
# for each server, Check for missing full backups
#
Write-Host "2) FULL Backups Not Performed Recently" -BackgroundColor Blue -ForegroundColor White
$checkId = $checkId + 1
foreach ($svr in get-content $instanceListPath){
	#Write-Host $svr
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$svr;database=msdb;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT  d.Name , CAST(COALESCE(MAX(b.backup_finish_date), ' never ') AS VARCHAR(200)) AS LastBackup
	        FROM    master.sys.databases d
	                LEFT OUTER JOIN msdb.dbo.backupset b ON d.name = b.database_name  AND b.type = 'D'
	        WHERE   d.database_id <> 2
	        GROUP BY d.name
	        HAVING  MAX(b.backup_finish_date) <= DATEADD(dd, -1 * $NbrDaysWithoutBackup, GETDATE())"
	$cmd.CommandText = $sql
	$da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	[void]$da.fill($ds)
	$cn.Close()
	$ds.Tables[0] | foreach {
		#Write-Host "database " $_.Name
		$tmpDetails = "" + $_.Name + " database last backed up: " + $_.LastBackup
		# log it
		InsertDailyHealthCheck $svr $checkId $priority "Backup" "FULL Backups Not Performed Recently" $tmpDetails 

	}
	#$ds.Tables[0] | ft
} # end for each server

# Check 3
# for each server, Check FULL Backups not performed ever
#
Write-Host "3) FULL Backups not performed ever" -BackgroundColor Blue -ForegroundColor White
$checkId = $checkId + 1
foreach ($svr in get-content $instanceListPath){
	#Write-Host $svr
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$svr;database=msdb;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT (d.NAME + ' database created on ' + CONVERT(NVARCHAR(10), d.create_date, 101) + ' never backed up.') AS Details
			FROM master.sys.databases d
			LEFT OUTER JOIN msdb.dbo.backupset bkst ON d.NAME = bkst.database_name
				AND bkst.type = 'D'
			WHERE d.database_id <> 2
				AND bkst.backup_finish_date IS NULL"
	$cmd.CommandText = $sql
	$da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	[void]$da.fill($ds)
	$cn.Close()
	$ds.Tables[0] | foreach {
		#Write-Host "database " $_.Name
		$tmpDetails = $_.Details
		# log it
		InsertDailyHealthCheck $svr $checkId $priority "Backup" "FULL Backup never performed" $tmpDetails 

	}
	#$ds.Tables[0] | ft
} # end for each server

# Check 4
# for each server, Check for databases in FULL recovery mode with no LOG backups
#
Write-Host "4) FULL recovery mode with no LOG backups" -BackgroundColor Blue -ForegroundColor White
$checkId = $checkId + 1
foreach ($svr in get-content $instanceListPath){
	#Write-Host $svr
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$svr;database=msdb;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT DISTINCT ((d.NAME COLLATE database_default) + ' database is in ' + d.recovery_model_desc + ' recovery mode but has not had a log backup in the last $NbrDaysWithoutLogBackup days. Last LOG backup on ' + ISNULL(CONVERT(varchar(10), bkst.last_backup_date,101), 'NEVER')) AS Details
			FROM master.sys.databases d 
			LEFT OUTER JOIN
			(
				SELECT database_name, MAX(backup_finish_date) as last_backup_date
				FROM msdb.dbo.backupset
				WHERE [type] = 'L'
				GROUP BY database_name
				HAVING MAX(backup_finish_date) >= DATEADD(dd, -1*$NbrDaysWithoutLogBackup, DATEADD(dd, DATEDIFF(dd, 0, getdate()), 0))
			) bkst ON
			 d.name = bkst.database_name
			WHERE d.recovery_model IN (1,2) --full, bulk logged
				AND d.database_id NOT IN (2,3) --tempdb, model
				AND d.source_database_id IS NULL
				AND (last_backup_date <= DATEADD(dd, -1*$NbrDaysWithoutLogBackup, getdate())
				OR last_backup_date IS NULL)"
	$cmd.CommandText = $sql
	$da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	[void]$da.fill($ds)
	$cn.Close()
	$ds.Tables[0] | foreach {
		#Write-Host "database " $_.Name
		$tmpDetails = $_.Details
		# log it
		InsertDailyHealthCheck $svr $checkId $priority "Backup" "Full Recovery Mode w/o Log Backups" $tmpDetails 

	}
	#$ds.Tables[0] | ft
} # end for each server

# Check 5
# for each server, Check for SQL Agent operators (notification) being configured
#    (issue if SQL Server is reinstalled, user db's restored, but not sys db's)
#
Write-Host "5) Reliability - No SQL Server Agent operators (emails) have been configured" -BackgroundColor Blue -ForegroundColor White
$checkId = $checkId + 1
$priority = 50
foreach ($svr in get-content $instanceListPath){
	#Write-Host $svr
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$svr;database=msdb;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT ('No SQL Server Agent Notification operators configured. Make sure to configure Mail profile, enable in Agent properties Alert System, create an Operator, and restart Agent service') AS Details
			WHERE NOT EXISTS (
				SELECT *
				FROM msdb.dbo.sysoperators
				WHERE enabled = 1)
				AND 0 = CASE WHEN @@Version LIKE '%Express Edition%' THEN 1
				ELSE 0	END"
	$cmd.CommandText = $sql
    try{
        $da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	    [void]$da.fill($ds)
        $ds.Tables[0] | foreach {
		        #Write-Host "database " $_.Name
		        $tmpDetails = $_.Details
               
		        # log it
		        InsertDailyHealthCheck $svr $checkId $priority "Reliability" "No Operators Configured/Enabled" $tmpDetails 

	     }
    } catch [system.exception]{
        $sqlEx = "5) Reliability: Caught exception, possibly a SQL Express instance with no SQL Agent. "
        InsertDailyHealthCheck $svr $checkId $priority "Reliability" "No Operators Configured/Enabled" $sqlEx 
    } finally{
        $cn.Close()
    }
	
	
	#$ds.Tables[0] | ft
} # end for each server
#-------------------------------------------
# create the report and send the email
#-------------------------------------------
if ($useThisScriptToSendEmail -eq $true){
	$runDate = Get-Date -Format 'yyyy-MM-dd'
	$ds = new-object System.Data.DataSet
	$cn = new-object System.Data.SqlClient.SqlConnection "server=$loggingServer;database=$loggingDatabase;Integrated Security=sspi"
	$cn.Open()
	$cmd = $cn.CreateCommand()
	$sql = "SELECT  [ID]
      ,[Server]
      ,[CheckID]
      ,[Priority]
      ,[FindingsGroup]
      ,[Finding]
      ,[Details]
      ,[RunDate]
      ,[CreatedOn]
	  FROM [dbo].[SqlServerDailyHealthCheck]
	  WHERE RunDate = @RunDate 
	  ORDER BY Server, CheckID"
	$cmd.CommandText = $sql
	$cmd.CommandTimeout = 0
	$cmd.Parameters.AddWithValue("@RunDate", $runDate) | Out-Null
	$da = New-Object System.Data.SqlClient.SqlDataAdapter($cmd)
	[void]$da.fill($ds) 
	$cn.Close()
	#
	# build the html body of the exception report for the email message
	#
	$sb = New-Object system.Text.StringBuilder
	$sb.AppendLine("<html><head>") | Out-Null
	# add some styles to pretty it up a bit
    $sb.AppendLine("<style type='text/css'>") | Out-Null
    $sb.AppendLine("body { font-family:Arial, Tahoma, 'Microsoft Sans Serif'; font-size: 10px; }") | Out-Null
    $sb.AppendLine("table { border-collapse:collapse } ") | Out-Null
    $sb.AppendLine("th, td { vertical-align:top; padding: 4px } ") | Out-Null
    $sb.AppendLine("table.grid td,table.grid th{ border: 1px solid #cccccc;  }") | Out-Null
    $sb.AppendLine("th { background-color: #f1f1f1 } ")| Out-Null
	# main body
	$sb.AppendLine("<body><table class='grid'><tr><th>Server</th><th>Check #</th><th>Type<th>Finding</th><th>Details</th></tr>") | Out-Null
	# for each row in the query add a table row
	$ds.Tables[0] | foreach {
		$sb.AppendLine("<tr>")| Out-Null
		$sb.Append("<td>") | Out-Null
		$sb.Append($_.Server) | Out-Null
		$sb.Append("</td>") | Out-Null
		$sb.Append("<td>") | Out-Null
		$sb.Append($_.CheckID) | Out-Null
		$sb.Append("</td>") | Out-Null
		$sb.Append("<td>") | Out-Null
		$sb.Append($_.FindingsGroup) | Out-Null
		$sb.Append("</td>") | Out-Null
		$sb.Append("<td>") | Out-Null
		$sb.Append($_.Finding) | Out-Null
		$sb.Append("</td>") | Out-Null
		$sb.Append("<td>") | Out-Null
		$sb.Append($_.Details) | Out-Null
		$sb.Append("</td>") | Out-Null
		#$tmpDetails = $_.Server + " run date: " + $_.RunDate + " " + $_.Finding
		$sb.AppendLine("</tr>")| Out-Null
	}
	$sb.AppendLine("</table></body></html>")| Out-Null
	Write-Host $sb.ToString()
	#
	# send the mail message
	#
	Send-MailMessage -From $emailFrom `
	-To $emailTo -Subject $emailSubject `
	-SmtpServer $emailServerHost -BodyAsHtml $sb.ToString()
} #end use this script to send email